
#include <stdlib.h>
#include <stdio.h>

/* 
  wavevote v0.0 - (C) by Bernhard Oemer (oemer@tph.tuwien.ac.at) 1998

  This is free software according to the GPL.

  BUILD: gcc wavevote.c -o wavevote

  USAGE: wavevote file1.wav file2.wav [...] 1>output.wav 2>log.txt
  Alle files must be cdda-takes of the same track and have to be of the same length.
  The interpolated output-file is sent to stdout, logs are sent to stderr.
*/ 


#define sample int		/* datatye for cdda-sample, must be sizeof 4 */
#define MAXFILES 16		/* max. no. of files in the cmd-line */
#define BUFLEN 256		/* length (in 4byte-samples) of the sync-buffer */
#define SYNCLEN 4		/* no. of misaliged values before sync is tried */
#define CMPOFFS (BUFLEN/4)	/* max. offset that can be resynced */
#define CMPLEN (BUFLEN/2)	/* no. of samples to be compared for resync */

/* wavevote falls back to interpolation is case the voting failes
   (unfortunatly this is all to often necessary with scratches).

   Interpolation takes place when a certain delta (difference between 2
   consecutive sample values) is execeeded. The delta is adaptive.
*/

#define DELTA 256		/* min. value of delta */
#define RESPONSE 32		/* speed of adapting delta to the avg. delta
                                   found in the data. Low values mean fast 
                                   adaption. */

/* Macros for converting the 4-byte-cdda-samples to integer amplitudes */

/* This can be activted, if data is in invers byte-order
#define swap(x) ((((x)>>8)&0x00ff)|(((x)<<8)&0xff00))
*/
#define swap(x) (x)
#define chop(x) (0xffff&((x)<-32768 ? -32768 : ((x)>32767 ? 32767 : (x))))
#define expand(x) (x & 0x8000 ? ((x) | 0xffff0000) : (x))
#define unpack(s,x,y) { x=expand(s&0xffff); y=expand(s>>16); }
#define pack(s,x,y) { s=(chop(x)|(chop(y)<<16)); }

FILE *f[MAXFILES];
int o[MAXFILES];

int c; /* resync count: set to SYNCLEN, decremented by misalignment, resync when 0 */
int n; /* number of input files */
int p; /* current position in the first wave-file, to which all other files are aligned */

/* Diagnostic variables */

int voterr=0;		/* No. of times when voting failed */
int interr=0;		/* No. of interpolations that took place */
int cmperr[MAXFILES];	/* Misreads per file */

sample b[MAXFILES][BUFLEN];	/* resync buffers for all data-files */
sample s[MAXFILES];		/* currently read sample for all data files */
int h[MAXFILES];		/* no. of votes h[i] for sample s[i] */
int delta=DELTA;		/* delta-threshold for scratch detection, see above */

/* Errorexit */

void err(char *s) {
  fprintf(stderr,"%s\n",s);
  exit(1);
}

/* Read from input file i */

sample rd(int i) {
  sample s;
  if(feof(f[i])) return 0;
  if(fread(&s,4,1,f[i])!=1) return 0;
  return s;
}

/* write to output */

void wr(sample s) {
  if(fwrite(&s,4,1,stdout)!=1) err("Write Error");
}

/* repostion read-pointer in file i by o samples */

void sk(int i,int o) {
  if(feof(f[i])) return;
  fseek(f[i],o*4,SEEK_CUR);
}

/* Print stats and exit */

void end() {
  int i;
  for(i=0;i<n;i++) {
    fclose(f[i]);
  }
  fprintf(stderr,"%d samples (%d:%2d), %d vote-errors, %d interpolations\n",
                  p,p/(60*44100),p/44100,voterr,interr);
  for(i=0;i<n;i++) {
    fprintf(stderr,"%2d: %d compare errors\n",i,cmperr[i]);
  };
  exit(0);
}

/* Resynchronize the readpointers in all input files to maximum 
   overlab with input file 0 */

void resync() {
  int i,j,k,l,m;
  c=SYNCLEN;		/* reset sync-counter */
  if(p>=CMPLEN) {	/* set read-pointer back */
    for(i=0;i<n;i++) {
      sk(i,-CMPLEN);
    }
  }
  for(i=0;i<n;i++) {	/* fill sync buffers */
    if(feof(f[i])) {
      for(j=0;j<BUFLEN;j++) b[i][j]=0;
    } else {
      l=fread(b[i],4,BUFLEN,f[i]);
      if(l<BUFLEN) {
        if(i==0) {
          for(j=CMPLEN;j<l;j++) wr(b[i][j]);
          end();
        } else {
          for(j=l;j<BUFLEN;j++) b[i][j]=0;
        };
      };
    };
  };
  if(p<CMPLEN) {	/* catch special case, when on the beginning of input file */
    for(i=0;i<CMPLEN;i++) wr(b[0][i]);
    p+=CMPLEN;
  };
  o[0]=0;
  for(i=1;i<n;i++) {	/* find offsets */
    m=-1;
    for(j=-CMPOFFS;j<CMPOFFS;j++) {
      l=0;
      for(k=CMPOFFS;k<CMPOFFS+CMPLEN;k++) {
        if(b[i][k+j]==b[0][k]) l++;
      };
      if(m<l) { o[i]=j; m=l; } 
    };
  };

  fprintf(stderr,"pos%9d   sync:",p);
  for(i=0;i<n;i++) {	/* apply offsets */
    sk(i,-CMPLEN+o[i]);
    fprintf(stderr," %2d:%+5d ",i,o[i]);
  };
  fprintf(stderr,"\n");
}

/* Main program */

main(int argc,char **argv) {
  int i,j,k,m;
  int d; 				/* actual delta including modifiers */
  int l,ll; 				/* last 2 sample values */
  int x,y,xx,yy,xxx,yyy,xs,ys,dx,dy;	/* vars for interpolation */
  int ipol;  				/* interpolation count */

  if(argc<3) err("At least 2 files are required.");
  n=argc-1;
  for(i=0;i<n;i++) {
    f[i]=fopen(argv[i+1],"r");
    fprintf(stderr,"%2d: opening %s\n",i,argv[i+1]);
    if(!f[i]) err("File open error.");
    cmperr[i]=0;
  }

/* read until first difference occures in one of the files;
   this is mainly for skipping headers and initial zero-samples */

  p=-1;
  while(1) {
    for(i=0;i<n;i++) {
      s[i]=rd(i);
    };
    p++;
    wr(s[0]);
    if(feof(f[0])) exit(0);
    for(i=0;i<n;i++) {
      for(j=0;j<i;j++) {
        if(s[i]!=s[j]) goto sync;
      }
    }
  }

sync:

  resync();

  ipol=ll=l=0;

  /* main loop */

  while(1) {
    for(i=0;i<n;i++) {
      s[i]=rd(i);
      h[i]=1;
    };
    if(feof(f[0])) end();
    
    /* Voting */
    
    for(i=0;i<n;i++) {
      for(j=0;j<i;j++) {
        if(s[i]==s[j]) {
          h[i]++; h[j]++;
        }
      }
    }
    m=0;
    for(i=1;i<n;i++) {
      if(h[i]>h[m]) m=i;
    }

    /* m contains the no. of file with best voting result
       s[m] contains the sample value voted for
       h[m] contains the number of files voted for s[m]
    */

    /* Vote-Error is triggered when no majority found */

    if(h[m]<n/2) {
      voterr++;
      fprintf(stderr,"pos%9d  error: voting failed\n",p);
      s[m]=l; /* Sample not to be trusted, temporarily set to last read sample */
    } else if (h[m]<2*n/3) { /* log decisions with <2/3 majority */
      fprintf(stderr,"pos%9d  %2d/%2d:",p,h[m],n);
      for(i=0;i<n;i++) fprintf(stderr," %08x%c",s[i],s[m]==s[i] ? '*' : ' ');
      fprintf(stderr,"\n");
    };

    /* Scratch Detection and Interpolation */

    unpack(s[m],x,y);	/* x and y stand for the left and right channel */
    xs=x;ys=y;
    unpack(l,xx,yy);	/* unpack last 2 sample values */
    unpack(ll,xxx,yyy);
    dx=abs(x-xx); dy=abs(y-yy);

    /* calculate new delta; this formula is very heuristic. For static
       dx=dy the limes is delta=4*dx+2*DELTA, thus a dx more than 4 time
       the observed avg. value (puls 2xextra minimum DELTA) is considerd 
       to be invalid. This can be temporarily changed by modifiers.
       Adaption speed is controlled by the value of RESPONSE.
    */

    d=delta=(((RESPONSE-1)*delta+2*(dx+dy+DELTA))/RESPONSE);

    /* apply modifiers */

    if(h[m]==n) d*=3;	/* temp. increase delta threshold if voting was unanimous */
    d*=(ipol+1);	/* temp. increase delta threshold if the last ipol
                           samples have allready been interpolated. This is to
                           avoid mislead interpolations cause by a sudden change
                           in volume */

    /* apply Interploation when voting failed or delta exceeds threshold.
       The interpol. ist linear, but damped by a factor of 1/2 - don't ask
       me why, but I found this to perform much better then strictly
       linear interpolation */

    if(h[m]<n/2 || dx>d) x=xx+(xx-xxx)/2;
    if(h[m]<n/2 || dy>d) y=yy+(yy-yyy)/2;

    /* log and update ipol-count if interpolation has taken place */

    if(x!=xs || y!=ys) {
      pack(s[m],x,y);
      interr++;
      fprintf(stderr,"pos%9d interp: %6d,%6d -> %6d,%6d (old: %6d,%6d, d=%d)\n"
    		    ,p,xs,ys,x,y,xx,yy,delta);
      ipol++;
    } else {
      ipol=0;
    }

    /* check if resync is necessary */

    if(h[m]<n) {
      c--;
      if(c<=0) resync();
    } else {
      c=SYNCLEN;
    };

    /* Bookkeeping of error for each input file */

    for(i=0;i<n;i++) {
      if(s[i]!=s[m]) cmperr[i]++;
    };

    /* Finally write sample to stdout */

    wr(s[m]); p++; ll=l; l=s[m];
  }
}
